;; Generated package description from doc-toc.el  -*- no-byte-compile: t -*-
(define-package "doc-toc" "1.0.0.20220926.110900" "Manage outlines/table of contents of pdf and djvu documents" '((emacs "26.1")) :commit "b4bb748aa303517cef1e1ce4c08047664704dbd5" :authors '(("Daniel Laurens Nicolai" . "dalanicolai@gmail.com")) :maintainer '("Daniel Laurens Nicolai" . "dalanicolai@gmail.com") :keywords '("tools" "outlines" "convenience") :url "https://github.com/dalanicolai/doc-tools-toc")
